part of ac_dart.host_service.service;

class _AcHostServiceAuthentication {
  final AcHostRegistrationRepository _hostRegistrationRepository;
  final AddonHost _addonHost;
  final Set<JsonWebAlgorithm> _supportedAlgorithms;

  _AcHostServiceAuthentication(
      AcHostRegistrationRepository hostRegistrationRepository,
      AddonHost addonHost, {Set<JsonWebAlgorithm> supportedAlgorithms})
      : this._hostRegistrationRepository = checkNotNull(
          hostRegistrationRepository),
        this._addonHost = checkNotNull(addonHost),
        this._supportedAlgorithms = supportedAlgorithms != null
            ? supportedAlgorithms
            : new Set.from([JsonWebAlgorithm.HS256]);

  Future<Either<AuthenticationFailure, AcSessionContext>> authenticateHostRequest(
      String jwtToken, Uri uri, String method) {
    final JsonWebToken jwt = decodeProductHostToken(jwtToken);

    final claimSet = jwt.claimSet;

    return _authenticate(jwt, claimSet.issuer, (AcHostRegistration hostRego) =>
        _createValidationContext(uri, method, hostRego));
  }

  JwtValidationContext _createValidationContext(
      Uri uri, String method, AcHostRegistration hostRego) {
    final JwaSignatureContext signatureContext =
        new JwaSymmetricKeySignatureContext(hostRego.sharedSecret);

    final claimSetValidationContext = new ProductHostClaimSetValidationContext(
        uri, method, contextPath: _addonHost.baseUrl.path);

    return new JwtValidationContext(
        signatureContext, claimSetValidationContext);
  }

  String createAddonSessionToken(String hostKey, String user) {
    if (_addonHost.addonSecret is None) {
      throw new ArgumentError('must configure a secret for the addon to '
          'create session tokens');
    }
    return session_token.createAddonSessionToken(
        _addonHost.addonSecret.get(), _addonHost.key, hostKey, user);
  }

  Future<Either<AuthenticationFailure, AcSessionContext>> authenticateAddonSessionToken(
      String jwtToken) {
    final JsonWebToken jwt = decodeAddonSessionToken(jwtToken);
    final AddonSessionClaimSet claimSet = jwt.claimSet;

    final issuer = claimSet.issuer;

    if (issuer != _addonHost.key) {
      return new Future.value(new Left(new AuthenticationFailure.singleError(
          HttpStatus.UNAUTHORIZED, "The token was not issued by this addon")));
    }

    return _authenticate(jwt, claimSet.productHostKey,
        (_) => new JwtValidationContext.withSharedSecret(
            _addonHost.addonSecret.get()));
  }

  Future<Either<AuthenticationFailure, AcSessionContext>> _authenticate(
      JsonWebToken<OpenIdJwtClaimSet> jwt, String issuer,
      JwtValidationContext createValidationContext(
          AcHostRegistration hostRego)) {
    final header = (jwt as JsonWebSignature).header;
    if (!_supportedAlgorithms.contains(header.algorithm)) {
      return new Future.value(new Left(new AuthenticationFailure.singleError(
          HttpStatus.UNAUTHORIZED, "unsupported algorithm")));
    }

    return _hostRegistrationRepository
        .findRegistrationByHostKey(issuer)
        .then((hostRegoOpt) {
      return hostRegoOpt.map((hostRego) {
        _log.finest('hostRego is $hostRego');

        Set<ConstraintViolation> violations =
            jwt.validate(createValidationContext(hostRego));

        return _createResult(violations, jwt.claimSet, hostRego);
      }).getOrElse(() => new Left(new AuthenticationFailure.singleError(
          HttpStatus.FORBIDDEN, "unknown host")));
    });
  }

  Either<AuthenticationFailure, AcSessionContext> _createResult(
      Set<ConstraintViolation> violations, OpenIdJwtClaimSet claimSet,
      AcHostRegistration hostRego) {
    if (violations.isEmpty) {
      return new Right(
          new AcSessionContext(claimSet.subject, hostRego.hostDetails.key));
    } else {
      return new Left(new AuthenticationFailure(HttpStatus.UNAUTHORIZED,
          new ConstraintViolations('Authentication failed', violations)));
    }
  }
}

class AuthenticationFailure {
  final int statusCode;
  final ConstraintViolations constraintViolations;

  String get message =>
      'Authentication Failed: ${constraintViolations.summaryMessage}';

  AuthenticationFailure(this.statusCode, this.constraintViolations);

  AuthenticationFailure.singleError(int statusCode, String message) : this(
          statusCode, new ConstraintViolations('Authentication failed',
              new Set()..add(new ConstraintViolation(message))));
}

class AcSessionContext {
  final String user;
  final String hostKey;

  AcSessionContext(this.user, this.hostKey);
}
